@@ -26,7 +26,7 @@ describe("basic index building", function() {
     albumArtistName: "Anberlin",
   });
 
-  library.rebuild();
+  library.rebuildTracks();
 
   it("trackTable", function() {
     var track = library.trackTable["Anberlin/Never Take Friendship Personal/08. The Feel Good Drag.mp3"];
@@ -128,7 +128,7 @@ describe("compilation album", function() {
     trackCount: 25,
   });
 
-  library.rebuild();
+  library.rebuildTracks();
 
   it("filed in various artists", function() {
     assert.strictEqual(library.albumList.length, 2);
@@ -177,7 +177,7 @@ describe("tracks from same album missing year metadata", function() {
     track: 3,
   });
 
-  library.rebuild();
+  library.rebuildTracks();
 
   it("still knows they're in the same album", function() {
     assert.strictEqual(library.albumList.length, 1);
@@ -209,7 +209,7 @@ describe("different albums with same name", function() {
     track: 2,
   });
 
-  library.rebuild();
+  library.rebuildTracks();
 
   it("detects that they are different", function() {
     assert.strictEqual(library.albumList.length, 2);
@@ -242,7 +242,7 @@ describe("album with a few tracks by different artists", function() {
   });
 
 
-  library.rebuild();
+  library.rebuildTracks();
 
   it("only creates one album", function() {
     assert.strictEqual(library.albumList.length, 1);
@@ -271,7 +271,7 @@ describe("album by an artist", function() {
   });
 
 
-  library.rebuild();
+  library.rebuildTracks();
 
   it("should be filed under the artist", function() {
     assert.strictEqual(library.artistList.length, 1);
@@ -333,7 +333,7 @@ describe("album by an artist", function() {
     trackCount: 25,
   });
 
-  library.rebuild();
+  library.rebuildTracks();
 
   it("sorts by disc before track", function() {
     assert.strictEqual(library.albumList[0].trackList[0].name, "No News Is Good News");
@@ -362,7 +362,7 @@ describe("album artist with no album", function() {
     genre: 'Rock',
   });
 
-  library.rebuild();
+  library.rebuildTracks();
 
   it("shouldn't be various artists", function() {
     assert.notStrictEqual(library.trackTable[id].albumArtistName, "Various Artists");
@@ -388,7 +388,7 @@ describe("unknown artist, unknown album", function() {
     albumName: '',
   });
 
-  library.rebuild();
+  library.rebuildTracks();
   var results = library;
 
   it("should be put into the same album", check);
@@ -438,7 +438,7 @@ describe("album with album artist", function() {
     albumName: 'Escala',
     track: 3,
   });
-  library.rebuild();
+  library.rebuildTracks();
 
   it("shouldn't be various artists", function() {
     assert.strictEqual(library.trackTable[id1].albumArtistName, "Escala");
@@ -447,19 +447,97 @@ describe("album with album artist", function() {
   });
 });
 
-describe("parseQueryIntoTerms", function() {
+
+describe("label management", function() {
+  var library = new MusicLibraryIndex();
+
+  library.addLabel({
+    id: "wrong_id",
+    name: "wrong",
+  });
+  library.rebuildLabels();
+  library.clearLabels();
+
+  library.addLabel({
+    id: "techno_id",
+    name: "techno",
+  });
+  library.addLabel({
+    id: "jazz_id",
+    name: "jazz",
+  });
+  library.rebuildLabels();
+
+  it("clearLabels, addLabel", function() {
+    assert.strictEqual(library.labelList[0].name, "jazz");
+    assert.strictEqual(library.labelList[1].name, "techno");
+  });
+});
+
+describe("parseQuery", function() {
+  var library = new MusicLibraryIndex();
   it("works", function() {
-    assert.deepEqual(MusicLibraryIndex.parseQueryIntoTerms(""), []);
-    assert.deepEqual(MusicLibraryIndex.parseQueryIntoTerms("a"), ["a"]);
-    assert.deepEqual(MusicLibraryIndex.parseQueryIntoTerms(" a   b "), ["a", "b"]);
-    assert.deepEqual(MusicLibraryIndex.parseQueryIntoTerms("\"a  b\""), ["a  b"]);
-    assert.deepEqual(MusicLibraryIndex.parseQueryIntoTerms("\"a  b\\\" c\""), ["a  b\" c"]);
-    assert.deepEqual(MusicLibraryIndex.parseQueryIntoTerms("\\\"a  b\""), ["\"a", "b\""]);
-    assert.deepEqual(MusicLibraryIndex.parseQueryIntoTerms("\""), ["\""]);
-    assert.deepEqual(MusicLibraryIndex.parseQueryIntoTerms("\\"), ["\\"]);
-    assert.deepEqual(MusicLibraryIndex.parseQueryIntoTerms("\"\""), []);
-    assert.deepEqual(MusicLibraryIndex.parseQueryIntoTerms("a\" b\"c"), ["a\"", "b\"c"]);
-    assert.deepEqual(MusicLibraryIndex.parseQueryIntoTerms("ab cd e"), ["ab", "cd", "e"]);
+    assert.strictEqual(library.parseQuery('').toString(), '()');
+    assert.strictEqual(library.parseQuery('a').toString(), '(fuzzy "a")');
+    assert.strictEqual(library.parseQuery(' ab   cd ').toString(), '((fuzzy "ab") AND (fuzzy "cd"))');
+    assert.strictEqual(library.parseQuery('"a  b"').toString(), '(exact "a  b")');
+    assert.strictEqual(library.parseQuery('"a  b\\" c"').toString(), '(exact "a  b\\\" c")');
+    assert.strictEqual(library.parseQuery('\\"a  b"').toString(), '((fuzzy "\\\\\\"a") AND (fuzzy "b\\""))');
+    assert.strictEqual(library.parseQuery('"').toString(), '(fuzzy "\\\"")');
+    assert.strictEqual(library.parseQuery('\\').toString(), '(fuzzy "\\\\")');
+    assert.strictEqual(library.parseQuery('""').toString(), '()');
+    assert.strictEqual(library.parseQuery('a" b"c').toString(), '((fuzzy "a\\\"") AND (fuzzy "b\\\"c"))');
+
+    assert.strictEqual(library.parseQuery('not:A').toString(), '(not (fuzzy "a"))');
+    assert.strictEqual(library.parseQuery('not:"A"').toString(), '(not (exact "A"))');
+    assert.strictEqual(library.parseQuery('not:(a b)').toString(), '(not ((fuzzy "a") AND (fuzzy "b")))');
+    assert.strictEqual(library.parseQuery('not:(a)').toString(), '(not (fuzzy "a"))');
+    assert.strictEqual(library.parseQuery('not:not:a').toString(), '(not (not (fuzzy "a")))');
+    assert.strictEqual(library.parseQuery('not:').toString(), '(fuzzy "not:")');
+    assert.strictEqual(library.parseQuery('not: a').toString(), '((fuzzy "not:") AND (fuzzy "a"))');
+    assert.strictEqual(library.parseQuery('not:)a').toString(), '((fuzzy "not:)") AND (fuzzy "a"))');
+
+    assert.strictEqual(library.parseQuery('or:()').toString(), '()');
+    assert.strictEqual(library.parseQuery('or:(' ).toString(), '()');
+    assert.strictEqual(library.parseQuery('or:').toString(), '(fuzzy "or:")');
+    assert.strictEqual(library.parseQuery('or:(a)').toString(), '(fuzzy "a")');
+    assert.strictEqual(library.parseQuery('or:(a' ).toString(), '(fuzzy "a")');
+    assert.strictEqual(library.parseQuery('or:(a b)').toString(), '((fuzzy "a") OR (fuzzy "b"))');
+    assert.strictEqual(library.parseQuery('or:(a b' ).toString(), '((fuzzy "a") OR (fuzzy "b"))');
+    assert.strictEqual(library.parseQuery('or:(a (b c))').toString(), '((fuzzy "a") OR ((fuzzy "b") AND (fuzzy "c")))');
+    assert.strictEqual(library.parseQuery('or:((a b) c)').toString(), '(((fuzzy "a") AND (fuzzy "b")) OR (fuzzy "c"))');
+  });
+});
+
+describe("parseQuery with labels", function() {
+  var library = new MusicLibraryIndex();
+  library.addLabel({
+    id: "techno_id",
+    name: "techno",
+  });
+  library.addLabel({
+    id: "jazz_id",
+    name: "jazz music",
+  });
+  library.addLabel({
+    id: "not_id",
+    name: "not:",
+  });
+  library.rebuildLabels();
+
+  it("works", function() {
+    assert.strictEqual(library.parseQuery('label:techno').toString(), '(label "techno_id")');
+    assert.strictEqual(library.parseQuery('not:label:techno').toString(), '(not (label "techno_id"))');
+    assert.strictEqual(library.parseQuery('label:asdf').toString(), '(label <none>)');
+    assert.strictEqual(library.parseQuery('label:Techno').toString(), '(label <none>)');
+    assert.strictEqual(library.parseQuery('label:"jazz music"').toString(), '(label "jazz_id")');
+    assert.strictEqual(library.parseQuery('label:jazz music').toString(), '((label <none>) AND (fuzzy "music"))');
+    assert.strictEqual(library.parseQuery('label:""').toString(), '(label <none>)');
+
+    assert.strictEqual(library.parseQuery('label:').toString(), '(fuzzy "label:")');
+    assert.strictEqual(library.parseQuery('label: ').toString(), '(fuzzy "label:")');
+    assert.strictEqual(library.parseQuery('label:not:').toString(), '(label "not_id")');
+    assert.strictEqual(library.parseQuery('not:(this label:)').toString(), '(not ((fuzzy "this") AND (fuzzy "label:")))');
   });
 });
 
@@ -480,6 +558,19 @@ describe("searching with quoted seach terms", function() {
     albumName: "The Weekend Challenge #3",
   });
 
+  library.addTrack({
+    key: "v7zwEPLs",
+    name: "Mista veri pakenee",
+    artistName: "Turmion Katilot (no diacritics)",
+    albumName: "Pirun nyrkki",
+  });
+  library.addTrack({
+    key: "sobHcy0I",
+    name: "Mistä veri pakenee",
+    artistName: "Turmion Kätilöt (with diacritics)",
+    albumName: "Pirun nyrkki",
+  });
+
   var literalQuoteKey = "G5FqXeJZ";
   library.addTrack({
     key: literalQuoteKey,
@@ -496,19 +587,30 @@ describe("searching with quoted seach terms", function() {
     albumName: "literalBackslash",
   });
 
-  library.rebuild();
+  library.rebuildTracks();
 
   it("single search term returns both", function() {
     var results = library.search("aka aka");
     assert.strictEqual(results.artistList.length, 2);
   });
 
-  it("quoted search term", function() {
-    var results = library.search("\"aka aka\"");
+  it("quoted search term is case sensitive", function() {
+    assert.strictEqual(library.search("\"andy\"").artistList.length, 0);
+    assert.strictEqual(library.search("\"ANDY\"").artistList.length, 0);
+    assert.strictEqual(library.search("\"Andy\"").artistList.length, 1);
+  });
+
+  it("quoted search terms include spaces", function() {
+    var results = library.search("\"AKA AKA\"");
     assert.strictEqual(results.artistList.length, 1);
     assert.strictEqual(results.artistList[0].name, "AKA AKA & Thalstroem");
   });
 
+  it("quoted search terms preserve diacritics", function() {
+    assert.strictEqual(library.search("Mistä").artistList.length, 2);
+    assert.strictEqual(library.search("\"Mistä\"").artistList.length, 1);
+  });
+
   it("matches a song with a literal quote", function() {
     var results = library.search("\"");
     assert.strictEqual(results.albumList.length, 1);
@@ -522,3 +624,66 @@ describe("searching with quoted seach terms", function() {
     assert.strictEqual(results.albumList[0].trackList[0].key, literalBackslashKey);
   });
 });
+
+describe("searching with expressions", function() {
+  var library = new MusicLibraryIndex();
+  library.addLabel({
+    id: "techno_id",
+    name: "techno",
+  });
+  library.addLabel({
+    id: "jazz_id",
+    name: "jazz",
+  });
+  library.rebuildLabels();
+
+  library.addTrack({
+    key: "fUPmxjMc",
+    name: "Été (Original Mix)",
+    artistName: "AKA AKA & Thalstroem",
+    albumName: "Varieté",
+    labels: {"jazz_id": 1},
+  });
+  library.addTrack({
+    key: "v7zwEPLs",
+    name: "Été (Remix)",
+    artistName: "Some Remixer",
+    albumName: "Varieté",
+    labels: {"techno_id": 1, "jazz_id": 1},
+  });
+  library.addTrack({
+    key: "zyGaKkrU",
+    name: "Tribute to Young Stroke AKA Young Muscle",
+    artistName: "Andy Kelley",
+    albumName: "The Weekend Challenge #3",
+  });
+  library.rebuildTracks();
+
+  it("'not:'", function() {
+    assert.strictEqual(library.search('not:andy').artistList.length, 2);
+    assert.strictEqual(library.search('not:remix variete').artistList.length, 1);
+    assert.strictEqual(library.search('not:"AKA AKA"').artistList.length, 2);
+    assert.strictEqual(library.search('not:"aka aka"').artistList.length, 3);
+    assert.strictEqual(library.search('not:(aka young)').artistList.length, 2);
+    assert.strictEqual(library.search('not:').artistList.length, 0);
+  });
+
+  it("'label:'", function() {
+    assert.strictEqual(library.search('label:techno').artistList.length, 1);
+    assert.strictEqual(library.search('label:jazz').artistList.length, 2);
+    assert.strictEqual(library.search('not:label:techno').artistList.length, 2);
+    assert.strictEqual(library.search('"label:techno"').artistList.length, 0);
+    assert.strictEqual(library.search('not:"label:techno"').artistList.length, 3);
+    assert.strictEqual(library.search('label:wrong').artistList.length, 0);
+    assert.strictEqual(library.search('not:label:wrong').artistList.length, 3);
+
+    assert.strictEqual(library.search('not:(label:jazz not:label:techno)').artistList.length, 2);
+  });
+
+  it("'or:'", function() {
+    assert.strictEqual(library.search('or:()').artistList.length, 0);
+    assert.strictEqual(library.search('or:(aka)').artistList.length, 2);
+    assert.strictEqual(library.search('or:(variete)').artistList.length, 2);
+    assert.strictEqual(library.search('or:(label:techno not:label:jazz)').artistList.length, 2);
+  });
+});
